/******************************************************************************************************************

   Column reordering for tables in pg_dump 
   
   1. Initialization: reads from a file with a full qualified column per line, format  "schema.tablename.colname" 
   (accepts comments beginning with # , discards blanks, ignores blank lines, rejects non-ascii chars)
   Only required are the columns to be reordered, the omitted columns will be placed after the specified ones.
    
   Hint: The following query produces all the original columns in the desired format (uncoment to get the type as a comment)
   
   SELECT namesp.nspname || '.' || t.relname || '.' || atr.attname
    -- || ' # ' || atr_type.typname  -- optional
    FROM pg_attribute atr,  pg_class t, pg_type atr_type, pg_namespace namesp
    WHERE  atr.attnum>0 and atr.attisdropped = 'f' and atr.attrelid = t.oid 
    and atr.atttypid = atr_type.oid and t.relnamespace = namesp.oid and t.relkind = 'r' 
    and namesp.nspname not like 'pg_%' and namesp.nspname != 'information_schema'
    ORDER BY namesp.nspname,t.relname,atr.attnum;
 
   2. Hook for pg_dump: the  getAttrPermutation() function is used for the CREATE TABLE dump
   (might be also used in dumpTableData() but does not seem to be be necessary at all)
   The ordering file is passed to pg_dump as option -r/--reorder-columns
   
   TODO: - should accept non-ascii chars ? 
         - should panic if used with version < 7.3 (copy to does not support explicit column ordering!)
         - man page for pg_dump 
         - only tested with pg 8.3, pg_dump, default format (test/implement other versions, formats, commands ? pg_dumpall ?)
  
   Hernn J. Gonzlez nov 2008
   
*********************************************************************************************************************************/


#include "pg_dump.h"

int initAttrsReorder(const char *filename);
int * getAttrPermutation(const TableInfo *ti);
int showAttrsReorderForDebug();

typedef struct {
  char **attrsFullQual;  /* attributes to reorder, full cualified names: "schema.tablename.colname" */
  int size;  /* size of attrsFullQual array */
  int naloc; /* size of preallocated elements (no bytes!) in attrsFullQual*/
} _attrsToReorder;
static _attrsToReorder attrsToReorder;
static int initialized = 0; /* flag */

static int oldPos(char * fullattr,const TableInfo *ti);

#define MAX_LINE_LEN (64*4) /*  schema.tablename.colname (plus optional coment/type) should be enough */

/**
 *  Given tableInfo returns a permutation array
 *  Eg: perm = getAttrPermutation(tPersons)  with PERSONS ={NAME,AGE,EMAIL,ID} and the desired new order is PERSONS ={EMAIL,NAME,AGE,ID}
 *  then perm = [2,0,1,3] (to be understood: perm[0] = 2 implies that the first new column corresponds to the  column 2 in the original table
 *   (0 based index - does not treat specially dropped or inherited columns)
 *  Returns NULL if not initialized
 */
int * getAttrPermutation(const TableInfo *ti) {
   char prefix[MAX_LINE_LEN] ; /* namespace.table. */ 
   int  prefixlen; 
   int i,j,k;
   int *perm; 
   int *perminv; /* aux: inverse perm: index refers to old ordering, val=where has it been put in the new order */
   if(initialized==0)  { return NULL;} 
   perm = (int *)malloc((sizeof (int))* (ti->numatts));    
   perminv = (int *)malloc((sizeof (int))* (ti->numatts)); 
   for (i=0;i<ti->numatts;i++) perm[i] = perminv[i] = -1;
   sprintf(prefix,"%s.%s.",ti->dobj.namespace->dobj.name,ti->dobj.name);
   prefixlen = strlen(prefix);
   /* loop over all columns of this table */
   i = 0; /* index in perm[] */
   for(k=0;k<attrsToReorder.size;k++) {
   	if( strncmp(prefix,attrsToReorder.attrsFullQual[k],prefixlen)!=0) continue;
   	j = oldPos(attrsToReorder.attrsFullQual[k],ti);
   	if( j<0 )              
   	  { fprintf(stderr,"Column reorder failed: Attribute '%s' no found in table ?\n",attrsToReorder.attrsFullQual[k]); exit(1); }
   	/* assert; this should never happen */
   	if(i >= ti->numatts || j >= ti->numatts) 
   	  { fprintf(stderr,"Column reorder failed: Panic processing %s (%d,%d)\n",attrsToReorder.attrsFullQual[k],i,j); exit(1); }
   	/* was already mapped ? error */
   	if( perminv[j] >= 0)    
   	  { fprintf(stderr,"Column reorder failed: Error processing %s (duplicated entry?)\n",attrsToReorder.attrsFullQual[k]); exit(1); }
   	perm[i] = j;
   	perminv[j]=i;
   	i++;
   }
   j = 0;
   /* for the remaining (no specified in reorder file) columns  */
   for(;i<ti->numatts;i++) {
   	/*find next unmapped column */
   	while(perminv[j]>=0) { 
   	   j++; 
   	   if(j >= ti->numatts)  { fprintf(stderr,"Column reorder failed: Panic processing (%d,%d) \n",i,j); exit(1); }
   	}
   	perm[i] = j;
   	perminv[j]=i;
   }
   /* paranoid assert: is it a valid permutation array ? */
   for(i=0;i<ti->numatts;i++) {
   	j = perm[i];
   	if(j<0 || j >= ti->numatts)  { fprintf(stderr,"Column reorder failed: Check 1 failed (%d,%d) \n",i,j); exit(1); }
   	if(perminv[j] != i)          { fprintf(stderr,"Column reorder failed: Check 2 failed (%d,%d) \n",i,j); exit(1); }
   }
   free(perminv);
 /* fprintf(stderr,"buscando prefix %s : [",prefix);	
   for(i=0;i<ti->numatts;i++) fprintf(stderr,"%d ",perm[i]);
   fprintf(stderr,"]\n"); */
   return perm;
}

/* auxiliar function: given a attrFullQual (scheme.table.attr) return the position in which is presently in the table (-1 : not found) */
static int oldPos(char * fullattr,const TableInfo *ti) {
     int pos;
     char *atr = rindex(fullattr,'.') + 1;	
     for( pos = 0 ; pos < ti->numatts ; pos++ ) {
     	if( strcmp( atr , ti->attnames[pos] ) == 0 ) 
     	   return pos;
     }
     return -1;/* not found */
}


/* Cleans and check the line from the file, and adds it (except if it is blank)
    Returns:  0=not added 1=added  -1=error */
static int addLine(char *line) {
   char *p;
   char *nueva;
   int contpoints=0;
   /* first pass : skip initial blanks and trim comments and trailing blanks*/
   while(*line != '\0' && isspace(*line)) 
      line++; 
   for(p = line;*p != '\0';p++) { 
   	if(*p == '#' || isspace(*p) ) {  *p = '\0';  break;}
   }
   if(strlen (line) == 0) return 0; /* nothing to do */
   for(p = line;*p != '\0';p++) { /* second pass: check */
   	if(*p == '.') {contpoints++; }
   	else if(! isalnum(*p) && *p != '_') {  fprintf(stderr,"invalid caracter  (%c)\n",*p); return -1; }
   	/* TODO: nonascii chars in attrib names? */
   }
   if(contpoints != 2) {  fprintf(stderr,"invalid line (%d != 2 dots )\n",contpoints); return -1; }
   /* ready to add */
   if(attrsToReorder.naloc < attrsToReorder.size +1 ) {
   	attrsToReorder.naloc += 256; /* increment number of attrs to prealoc */
   	attrsToReorder.attrsFullQual = realloc(attrsToReorder.attrsFullQual, (sizeof (char*)) * (attrsToReorder.naloc));
   }
   nueva = malloc( (sizeof (char))*(strlen(line)+1));
   strcpy(nueva,line);
   attrsToReorder.attrsFullQual[attrsToReorder.size] = nueva;
   attrsToReorder.size++;
   /*fprintf(stderr,"agregada  [%s] size=%d\n",line,  attrsToReorder.size);*/
   return 1;
}

/** 
    Init: load atrs to reorder from file 
    In case of error, prints msg to stderr and exit(1)
 */
int initAttrsReorder(const char *filename) { 
   char line[MAX_LINE_LEN];
   int contline;
   FILE* fp;
   int res;
   attrsToReorder.size=0; 
   attrsToReorder.naloc=0;
   attrsToReorder.attrsFullQual=NULL;
   fp = fopen( filename, "rt");
   if(fp == NULL) {
       fprintf(stderr,"Could not open file '%s' for reading\n",filename); 
       exit(1);
   }
   contline =0;
   while (fgets(line, MAX_LINE_LEN, fp) != NULL) {
       contline ++;
       if(strlen(line) == MAX_LINE_LEN-1) {
       	  fprintf(stderr,"Line too long (line %d)\n",contline); 
       	  exit(1);
       }
       res = addLine(line);
       if(res<0) {
       	  fprintf(stderr,"Fatal error  (line %d , file %s)\n",contline,filename); 
       	  exit(1);
       }
   }
   fclose(fp);
   initialized=1;   
   return 0;
}

/* just for debug: outputs list to stderr */
int showAttrsReorderForDebug() {
   int i;
   for(i=0;i<attrsToReorder.size;i++) {
   	fprintf(stderr,"[%s]\n",attrsToReorder.attrsFullQual[i]);
   }
   return 0;
}
